// ======================================================================
//
// MenuElement.cpp
// jeff grills
//
// copyright 1997 origin systems inc
//
// ======================================================================

#include "First.h"
#include "MenuElement.h"

#include "Bitmap.h"
#include "Game.h"
#include "Gl.h"
#include "Iff.h"
#include "MenuBitmap.h"
#include "MenuButton.h"
#include "MenuDragBar.h"
#include "MenuHorizontalSliderBar.h"
#include "MenuHover.h"
#include "MenuKey.h"
#include "MenuListBox.h"
#include "MenuRectangle.h"
#include "MenuSelector.h"
#include "MenuStaticListBox.h"
#include "MenuText.h"
#include "MenuTextEdit.h"
#include "MenuVerticalSliderBar.h"
#include "Mouse.h"
#include "StringFile.h"

#include <dinput.h>

// ======================================================================

const Tag TAG_HSB  = TAG3(H,S,B);
const Tag TAG_INT  = TAG3(I,N,T);
const Tag TAG_KEY  = TAG3(K,E,Y);
const Tag TAG_VSB  = TAG3(V,S,B);

const Tag TAG_ABT2 = TAG(A,B,T,2);
const Tag TAG_ABT3 = TAG(A,B,T,3);
const Tag TAG_DRAG = TAG(D,R,A,G);
const Tag TAG_DTAB = TAG(D,T,A,G);
const Tag TAG_ETAB = TAG(E,T,A,G);
const Tag TAG_HOVR = TAG(H,O,V,R);
const Tag TAG_IANS = TAG(I,A,N,S);
const Tag TAG_ISLT = TAG(I,S,L,T);
const Tag TAG_LANG = TAG(L,A,N,G);
const Tag TAG_LBOX = TAG(L,B,O,X);
const Tag TAG_SBIT = TAG(S,B,I,T);
const Tag TAG_SHRD = TAG(S,H,R,D);
const Tag TAG_SLBX = TAG(S,L,B,X);
const Tag TAG_SMOS = TAG(S,M,O,S);
const Tag TAG_SOND = TAG(S,O,N,D);
const Tag TAG_SUBS = TAG(S,U,B,S);
const Tag TAG_TINT = TAG(T,I,N,T);
const Tag TAG_TXED = TAG(T,X,E,D);
const Tag TAG_TXVR = TAG(T,X,V,R);

const int MAX_ELEMENT_TAB_STOPS = 64;

// ======================================================================

MenuElement *MenuElement::create(MenuElement *parent, int maxWidth, int maxHeight, Iff *iff)
{
	MenuElement *result;

	iff->enterForm();
	
	switch (iff->parentName())
	{
		case TAG_ELEM:
			result = NEW MenuElement(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_SBIT:
			result = NEW MenuBitmap(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_BUTT:
			result = NEW MenuButton(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_TOGG:
			result = NEW MenuToggle(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_SLCT:
			result = NEW MenuSelector(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_ISLT:
			result = NEW MenuIntSelector(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_ABIT:
			result = NEW MenuAnimatingBitmap(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_TEXT:
			result = NEW MenuText(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_HOVR:
			result = NEW MenuHover(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_TXED:
			result = NEW MenuTextEdit(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_RECT:
			result = NEW MenuRectangle(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_LBOX:
			result = NEW MenuListBox(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_SLBX:
			result = NEW MenuStaticListBox(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_VSB:
			result = NEW MenuVerticalSliderBar(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_HSB:
			result = NEW MenuHorizontalSliderBar(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_TXVR:
			result = NEW MenuTextVariable(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_TINT:
			result = NEW MenuTextInt(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_ABT2:
			result = NEW MenuAnimatingBitmap2(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_ABT3:
			result = NEW MenuAnimatingBitmap3(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_KEY:
			result = NEW MenuKey(parent, maxWidth, maxHeight, iff);
			break;

		case TAG_DRAG:
			result = NEW MenuDragBar(parent, maxWidth, maxHeight, iff);
			break;

		default:
			result = NULL;
			DEBUG_FATAL(true, ("LoadMenuElement bad token %08x", iff->parentName()));
	}

	iff->exitForm();

	return result;
}

// ======================================================================

MenuElement::MenuElement(MenuElement *newParent, int maxWidth, int maxHeight, Iff *iff)
: IoWin(NULL)
{
	menuId = NULL;
	menuState = MenuStateInactive;
	parent = newParent;
	menuResource = NULL;
	firstChild = NULL;
	nextSibling = NULL;

	tabStops = NULL;
	numTabStops = 0;
	curTabStop = 0;
	tabStop = TabOff;

	menuRect.x0 = 0;
	menuRect.y0 = 0;
	menuRect.x1 = maxWidth - 1;
	menuRect.y1 = maxHeight - 1;

	// allocate our tabStop array (if we're the menu's root)
	if (parent == NULL)
		tabStops = NEW Point[MAX_ELEMENT_TAB_STOPS];

	load(iff);
};

// ----------------------------------------------------------------------

MenuElement::~MenuElement(void)
{
	MenuElement   *me;
	MenuResource  *mr;

	parent   = NULL;

	delete []tabStops;
	tabStops = NULL;
	numTabStops = 0;
	curTabStop = 0;

	while (firstChild)
	{
		me = firstChild;
		firstChild = firstChild->nextSibling;
		delete me;
	}

	while (menuResource)
	{
		mr = menuResource;
		menuResource = menuResource->getNext();
		delete mr;
	}

	delete [] menuId;
	menuId   = NULL;
}

// ----------------------------------------------------------------------

void MenuElement::load(Iff *iff)
{

	if (iff)
	{
		if (iff->seekChunk(TAG_NAME, Iff::FlagOptional))
		{
			menuId = MEM_OWN(iff->read_char(Iff::ReadRest));
			iff->exitChunk();
		}

		if (iff->seekChunk(TAG_RECT, Iff::FlagOptional))
		{
			menuRect.x0 = iff->read_int32();
			menuRect.y0 = iff->read_int32();
			menuRect.x1 = iff->read_int32();
			menuRect.y1 = iff->read_int32();
			iff->exitChunk();
		}
		else
			if (iff->seekChunk(TAG_PTWH, Iff::FlagOptional))
			{
				int x0, y0, w, h;

				x0 = iff->read_int32();
				y0 = iff->read_int32();
				w  = iff->read_int32();
				h  = iff->read_int32();

				if (x0 < 0)
					x0 += menuRect.x1;
				if (y0 < 0)
					y0 += menuRect.y1;
				if (w < 0)
					w += (menuRect.x1 - menuRect.x0 + 1);
				if (h < 0)
					h += (menuRect.y1 - menuRect.y0 + 1);

				menuRect.x0 = x0;
				menuRect.y0 = y0;
				menuRect.x1 = menuRect.x0 + w;
				menuRect.y1 = menuRect.y0 + h;

				iff->exitChunk();
			}
			else
				if (iff->seekChunk(TAG_PT, Iff::FlagOptional))
				{
					menuRect.x0 = iff->read_int32();

					if (menuRect.x0 < 0)
						menuRect.x0 += menuRect.x1;

					menuRect.y0 = iff->read_int32();
					if (menuRect.y0 < 0)
						menuRect.y0 += menuRect.y1;

					menuRect.x1 = menuRect.x0;
					menuRect.y1 = menuRect.y0;

					iff->exitChunk();
				}

		if (iff->seekForm(TAG_SHRD, Iff::FlagOptional))
		{
			MenuResource *n, *last;

			last = NULL;
			while (!iff->endOfForm())
			{
				iff->enterForm();

					switch (iff->parentName())
					{
						case TAG_BITM:
							n = NEW MenuResourceBitmap(this, iff);
							break;

						case TAG_FONT:
							n = NEW MenuResourceFont(this, iff);
							break;

						case TAG_LANG:
							n = NEW MenuResourceLanguage(this, iff);
							break;

						case TAG_INT:
							n = NEW MenuResourceInteger(this, iff);
							break;

						case TAG_IANS:
							n = NEW MenuResourceIntegerAnimation(this, iff);
							break;

						case TAG_SOND:
							n = NEW MenuSound(this, iff);
							break;

						case TAG_SMOS:
							n = NEW MenuSetMouse(this, iff);
							break;

						default:
							DEBUG_FATAL(true, ("MenuElement::load bad shared type"));
							n = NULL;
					}

				iff->exitForm();

				if (last)
					last->setNext(n);
				else
					menuResource = n;

				last = n;
			}

			iff->exitForm();
		}

		if (iff->seekForm(TAG_SUBS, Iff::FlagOptional))
		{
			MenuElement *n, *last;

			last = NULL;
			while (!iff->endOfForm())
			{
				n = Game::createMenuElement(this, menuRect.x1-menuRect.x0+1, menuRect.y1-menuRect.y0+1, iff);

				if (last)
					last->nextSibling = n;
				else
					firstChild = n;

				last = n;
			}

			iff->exitForm();
		}

		// force this to be a tabstop?
		if (iff->seekChunk(TAG_ETAB, Iff::FlagOptional))
		{
			tabStop = TabForceOn;
			iff->exitChunk(); // 'ETAB'
		}

		// force this to be a tabstop?
		if (iff->seekChunk(TAG_DTAB, Iff::FlagOptional))
		{
			tabStop = TabForceOff;
			iff->exitChunk(); // 'DTAB'
		}
	}
}

// ----------------------------------------------------------------------

void MenuElement::getAbsoluteRect(Rect *rect) const
{
	MenuElement *m;

	// get the current rectangle
	*rect = menuRect;
	
	// walk up all the parents to get the abosolute position
	for (m = parent; m; m = m->parent)
	{
		rect->x0 += m->menuRect.x0;
		rect->y0 += m->menuRect.y0;
		rect->x1 += m->menuRect.x0;
		rect->y1 += m->menuRect.y0;
	}
}

// ----------------------------------------------------------------------

void MenuElement::addTabStop(int x, int y)
{
	MenuElement *p;
	int         i, insertHere;
	bool        done;

	if (parent != NULL)
	{
		// find the root
		for (p = parent; p->parent; p = p->parent)
			;

		p->addTabStop(x, y);
	}
	else
	{
		numTabStops++;
		DEBUG_FATAL(numTabStops >= MAX_ELEMENT_TAB_STOPS, ("MenuElement::addTabStop exceeded max number (%d >= %d)", numTabStops, MAX_ELEMENT_TAB_STOPS));

		// find the place of "insertion"
		// 1) sort by y
		// 2) sort by leftmost x
		// 3) if this entry already exists, then drop it
		for (insertHere = 0, done = false; insertHere < (numTabStops - 1) && !done; )
		{
			if (y < tabStops[insertHere].y)
				done = true;
			else
				if (y == tabStops[insertHere].y)
				{
					// the y's are equal, now sort by x
					if (x < tabStops[insertHere].x)
						done = true;
					else
						if (x == tabStops[insertHere].x)
						{
							// this entry already exists.  DON'T add it
							numTabStops--;
							return;
						}
				}

			if (!done)
				insertHere++;
		}

		// now make room for the new tabstop
		for (i = numTabStops - 1; i > insertHere; i--)
			tabStops[i] = tabStops[i - 1];

		tabStops[insertHere].x = x;
		tabStops[insertHere].y = y;
	}
}

// ----------------------------------------------------------------------

IoResult MenuElement::event(IoEventStruct *event)
{
	MenuElement *me;
	IoResult     result;
	bool         block;

	// -qq- if people don't play nicely, i'll have to copy the whole fucking struct.  please don't make me do that. i don't want to have to kill you.

	// adjust the location to be relative to this box
	event->mouseX -= menuRect.x0;
	event->mouseY -= menuRect.y0;

	for (block = false, me = firstChild; me; me = me->nextSibling)
	{
		result = me->event(event);
		DEBUG_FATAL(IoWinManager::isKill(result), ("MenuElement::event child returned kill"));
		if (IoWinManager::isBlock(result))
			block = true;
	}

	// put the location back the way we found it
	event->mouseX += menuRect.x0;
	event->mouseY += menuRect.y0;

	// check for tab spot increments/decrements
	if (parent == NULL)
	{
		if (event->type == IoEventKeyPress && event->ch == DIK_TAB)
		{
#if PLATFORM == PLATFORM_WIN95
			// check for "shift-tab"
			if (GetAsyncKeyState(VK_SHIFT))
				curTabStop--;
			else
				curTabStop++;
#else
			curTabStop++;
#endif
			if (curTabStop >= numTabStops)
				curTabStop = 0;
			if (curTabStop < 0)
				curTabStop = numTabStops - 1;

			Mouse::gotoXY(tabStops[curTabStop].x, tabStops[curTabStop].y);
		}
	}

	return (block ? IoResultBlock : IoResultPass);
}

// ----------------------------------------------------------------------

IoResult MenuElement::update(void)
{
	IoResult     result;
	bool         block;
	MenuElement  *me;
	MenuResource *mr;
	Rect         r;

	// take care of tab-stops, either reset the list, or add this tabstop to the list
	if (parent)
	{
		// if we're a tabstop then add our coordinate to the list
		if (tabStop == TabOn || tabStop == TabForceOn)
		{
			getAbsoluteRect(&r);
			addTabStop((r.x0 + r.x1) / 2, (r.y0 + r.y1) / 2);
		}
	}
	else
	{
		// if we made it here then this is the root of the menu.
		// reset the tabstops to zero and allow for the tabstops to be recalculated
		// NOTE: yes this is done every frame - because selectors can change what's viewed on the screen
		//       and you don't want to tab to a "non-tabspot"
		// -ck
		numTabStops = 0;
	}

	for (block = false, me = firstChild; me; me = me->nextSibling)
	{
		result = me->update();
		if (IoWinManager::isBlock(result))
			block = true;
	}

	for (mr = menuResource; mr; mr = mr->getNext())
		mr->update();

	return (block ? IoResultBlock : IoResultPass);
}

// ----------------------------------------------------------------------

void MenuElement::draw(void) const
{
#if 0
	if (DPrintFlag(DP_E_MENU))
	{
		MenuResource *mr;

		for (mr = menuResource; mr; mr = mr->getNext())
			if (mr->getType() == MenuResource::TypeInteger)
				DPRINT(("menu int '%s' value %d\n", mr->getId(), static_cast<MenuResourceInteger *>(mr)->getValue()));
	}
#endif

	menuDraw(0, 0);
}

// ----------------------------------------------------------------------

void MenuElement::menuEvent(int id, MenuElement *whom)
{
	MenuResource *mr;

	for (mr = menuResource; mr; mr = mr->getNext())
			mr->menuEvent(id, whom);

	if (parent)
		parent->menuEvent(id, whom);
}

// ----------------------------------------------------------------------

void MenuElement::menuDraw(int x, int y) const
{
	MenuElement *me;

	for (me = firstChild; me; me = me->nextSibling)
		me->menuDraw(x + menuRect.x0, y + menuRect.y0);

	if (DFlag(DF_E_MENU_BOXES))
	{
		Gl::setMaterial(MaterialList::red());
		Gl::rect(menuRect.x0+x, menuRect.y0+y, menuRect.x1+x, menuRect.y1+y);
	}
}

// ----------------------------------------------------------------------

MenuResource *MenuElement::findSharedResource(MenuResource::Type type, const char *id)
{
	MenuResource *mr, *result;

	// try to find the resource in my list
	for (mr = menuResource, result = NULL; !result && mr; mr = mr->getNext())
		result = mr->findSharedResource(type, id);

	// try to find the resouce in all my ancestors
	if (!result && parent)
		result = parent->findSharedResource(type, id);

	return result;
}

// ----------------------------------------------------------------------

MenuElement *MenuElement::findMenu(const char *findId, bool required)
{
	MenuElement *me, *result;

	UNREF(required);

	if (menuId && strcmp(menuId, findId) == 0)
		return this;

	for (result = NULL, me = firstChild; !result && me; me = me->nextSibling)
		result = me->findMenu(findId, false);

	DEBUG_FATAL(required && !result, ("MenuElement::findMenu '%s' not found", findId));

	return result;
}

// ----------------------------------------------------------------------

Bitmap *MenuElement::findSharedBitmap(const char *id)
{
	MenuResource *resource;

	resource = findSharedResource(MenuResource::TypeBitmap, id);
	DEBUG_FATAL(!resource, ("MenuElement::findSharedBitmap failed on '%s'\n", id));

	return static_cast<MenuResourceBitmap *>(resource)->getBitmap();
}

// ----------------------------------------------------------------------

Font *MenuElement::findSharedFont(const char *id)
{
	MenuResource *resource;

	resource = findSharedResource(MenuResource::TypeFont, id);
	DEBUG_FATAL(!resource, ("MenuElement::findSharedFont failed on '%s'\n", id));

	return static_cast<MenuResourceFont *>(resource)->getFont();
}

// ----------------------------------------------------------------------

StringFile *MenuElement::findSharedLanguage(const char *id)
{
	MenuResource *resource;

	resource = findSharedResource(MenuResource::TypeLanguage, id);
	DEBUG_FATAL(!resource, ("MenuElement::findSharedLanguage failed on '%s'\n", id));

	return static_cast<MenuResourceLanguage *>(resource)->getLanguage();
}

// ----------------------------------------------------------------------

MenuResourceInteger *MenuElement::findSharedInteger(const char *id)
{
	MenuResource *resource;

	resource = findSharedResource(MenuResource::TypeInteger, id);
	DEBUG_FATAL(!resource, ("MenuElement::findSharedInteger failed on '%s'\n", id));

	return static_cast<MenuResourceInteger *>(resource);
}

// ----------------------------------------------------------------------

MenuResourceIntegerAnimation *MenuElement::findSharedIntegerAnimation(const char *id)
{
	MenuResource *resource;

	resource = findSharedResource(MenuResource::TypeIntegerAnimation, id);
	DEBUG_FATAL(!resource, ("MenuElement::findSharedIntegerAnimation failed on '%s'\n", id));

	return static_cast<MenuResourceIntegerAnimation *>(resource);
}

// ======================================================================
